import json
from datetime import timedelta
from unittest.mock import MagicMock, patch, mock_open

import pytest
from tofupilot.utils import (
    validate_attachments,
    initialize_upload,
    upload_file,
    notify_server,
    parse_error_message,
    timedelta_to_iso,
    log_and_raise,
)
from tofupilot.constants import SECONDS_BEFORE_TIMEOUT


def test_validate_attachments():
    logger = MagicMock()
    attachments = ["file1.txt", "file2.jpg"]
    max_attachments = 2
    max_file_size = 5000

    with patch("os.path.getsize", return_value=4000):
        validate_attachments(logger, attachments, max_attachments, max_file_size)

    logger.info.assert_called_with("Validating attachments...")
    logger.error.assert_not_called()

    with pytest.raises(RuntimeError):
        with patch("os.path.getsize", return_value=6000):
            validate_attachments(
                logger,
                attachments,
                max_attachments,
                max_file_size,
            )

    logger.error.assert_called_with(
        "File size exceeds the maximum allowed size of 5000 bytes: file1.txt"
    )


def test_initialize_upload():
    headers = {"Authorization": "Bearer token"}
    base_url = "http://example.com"
    file_path = "file.txt"
    response_data = {"uploadUrl": "http://example.com/upload", "id": "123"}

    with patch("requests.post") as mock_post:
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = response_data

        upload_url, upload_id = initialize_upload(headers, base_url, file_path)

        assert upload_url == response_data["uploadUrl"]
        assert upload_id == response_data["id"]


def test_upload_file():
    upload_url = "http://example.com/upload"
    file_path = "file.txt"
    file_content = b"file content"

    with patch("builtins.open", mock_open(read_data=file_content)) as mock_file, patch(
        "requests.put"
    ) as mock_put, patch("mimetypes.guess_type", return_value=("text/plain", None)):
        mock_put.return_value.status_code = 200

        result = upload_file(upload_url, file_path)

        assert result is True
        mock_file.assert_called_once_with(file_path, "rb")
        mock_put.assert_called_with(
            upload_url,
            data=mock_file(),
            headers={"Content-Type": "text/plain"},
            timeout=SECONDS_BEFORE_TIMEOUT,
        )


def test_notify_server():
    headers = {"Authorization": "Bearer token"}
    base_url = "http://example.com"
    upload_id = "123"
    run_id = "456"

    with patch("requests.post") as mock_post:
        mock_post.return_value.status_code = 200

        result = notify_server(headers, base_url, upload_id, run_id)

        assert result is True
        mock_post.assert_called_with(
            f"{base_url}/uploads/sync",
            data=json.dumps({"upload_id": upload_id, "run_id": run_id}),
            headers=headers,
            timeout=SECONDS_BEFORE_TIMEOUT,
        )


def test_parse_error_message():
    response = MagicMock()
    response.json.return_value = {"error": {"message": "An error occurred"}}

    error_message = parse_error_message(response)
    assert error_message == "An error occurred"

    response.json.side_effect = ValueError
    response.text = "Some error text"
    error_message = parse_error_message(response)
    assert error_message == "HTTP error occurred: Some error text"


def test_timedelta_to_iso():
    td = timedelta(days=2, hours=3, minutes=4, seconds=5, microseconds=600000)
    iso_format = timedelta_to_iso(td)
    assert iso_format == "P2DT3H4M5.600000S"


def test_log_and_raise():
    logger = MagicMock()
    with pytest.raises(RuntimeError):
        log_and_raise(logger, "An error occurred")

    logger.error.assert_called_with("An error occurred")



from typing import Optional, TypedDict
from datetime import datetime, timedelta


class Step(TypedDict):
    name: str
    started_at: datetime
    duration: timedelta
    step_passed: bool
    measurement_unit: Optional[str]
    measurement_value: Optional[float]
    limit_low: Optional[float]
    limit_high: Optional[float]



import logging
import sys

# Define a custom log level for success messages
SUCCESS_LEVEL_NUM = 25
logging.addLevelName(SUCCESS_LEVEL_NUM, "SUCCESS")


def success(self, message, *args, **kws):
    """Log a success message."""
    if self.isEnabledFor(SUCCESS_LEVEL_NUM):
        self._log(SUCCESS_LEVEL_NUM, message, args, **kws)


logging.Logger.success = success


class CustomFormatter(logging.Formatter):
    """Custom formatter to add symbols and colors to the log messages."""

    reset_code = "\033[0m"

    format_dict = {
        logging.DEBUG: "\033[0;37m%(asctime)s - DEBUG: %(message)s"
        + reset_code,  # White
        logging.INFO: "\033[0;34m%(asctime)s - ℹ️ %(message)s"
        + reset_code,  # Blue with info symbol
        logging.WARNING: "\033[0;33m%(asctime)s - ⚠️ %(message)s"
        + reset_code,  # Yellow with warning symbol
        logging.ERROR: "\033[0;31m%(asctime)s - ❌ %(message)s"
        + reset_code,  # Red with cross mark
        logging.CRITICAL: "\033[1;41m%(asctime)s - 🚨 %(message)s"
        + reset_code,  # White on red background with alarm symbol
        SUCCESS_LEVEL_NUM: "\033[0;32m%(asctime)s - ✅ %(message)s"
        + reset_code,  # Green with checkmark
    }

    def format(self, record):
        """Format the specified record as text."""
        log_fmt = self.format_dict.get(record.levelno, self._fmt)
        formatter = logging.Formatter(log_fmt, datefmt="%Y-%m-%d %H:%M:%S")
        return formatter.format(record)


def setup_logger(log_level: int):
    """Set up the logger with a custom formatter and stream handler."""
    logger = logging.getLogger(__name__)
    logger.setLevel(log_level)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(log_level)
    handler.setFormatter(CustomFormatter())
    if not logger.handlers:
        logger.addHandler(handler)

    return logger



import subprocess
import os
import pytest


# This function discovers all Python scripts in the examples directory
def discover_example_scripts():
    examples_dir = "examples"
    scripts = []
    for filename in os.listdir(examples_dir):
        if filename.endswith(".py"):
            scripts.append(os.path.join(examples_dir, filename))
    return scripts


# Parametrize the test to run once for each script
@pytest.mark.parametrize("script_path", discover_example_scripts())
def test_example_script(script_path):
    result = subprocess.run(["python", script_path], capture_output=True, text=True)
    assert (
        result.returncode == 0
    ), f"Script {script_path} failed with error: {result.stderr}"



import logging
import os
from datetime import timedelta
from importlib.metadata import version
from typing import Dict, List, Optional

import requests

from .constants import (
    ENDPOINT,
    FILE_MAX_SIZE,
    CLIENT_MAX_ATTACHMENTS,
    SECONDS_BEFORE_TIMEOUT,
)
from .models import SubUnit, UnitUnderTest, Step
from .utils import (
    check_latest_version,
    handle_attachments,
    parse_error_message,
    setup_logger,
    timedelta_to_iso,
    datetime_to_iso,
    validate_attachments,
)


class TofuPilotClient:
    def __init__(self, api_key: Optional[str] = None, base_url: str = ENDPOINT):
        self._current_version = version("tofupilot")
        print_version_banner(self._current_version)  # Print the version banner
        self._api_key = api_key
        self._base_url = f"{base_url}/api/v1"
        self._headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._api_key}",
        }
        self._logger = setup_logger(logging.INFO)
        self._max_attachments = CLIENT_MAX_ATTACHMENTS
        self._max_file_size = FILE_MAX_SIZE
        check_latest_version(self._logger, self._current_version, "tofupilot")
        if api_key is None:
            api_key = os.environ.get("TOFUPILOT_API_KEY")
        if api_key is None:
            error = "Please provide an API key or set TOFUPILOT_API_KEY environment variable. For more information on how to find or generate a valid API key, visit https://docs.tofupilot.com/user-management#api-key."
            self._logger.error(error)
            raise Exception(error)

    def create_run(
        self,
        procedure_id: str,
        unit_under_test: UnitUnderTest,
        run_passed: bool,
        steps: Optional[List[Step]] = None,
        duration: timedelta = None,
        sub_units: Optional[List[SubUnit]] = None,
        report_variables: Optional[Dict[str, str]] = None,
        attachments: Optional[List[str]] = None,
    ) -> dict:
        """
        Creates a test run with the specified parameters and uploads it to the TofuPilot platform.
        [See API reference](https://docs.tofupilot.com/runs).

        Args:
            procedure_id (str): The unique identifier of the procedure to which the test run belongs.
            unit_under_test (UnitUnderTest): The unit being tested.
            run_passed (bool): Boolean indicating whether the test run was successful.
            duration (timedelta, optional): The duration of the test run. Default is None.
            steps (Optional[List[Step]], optional): [A list of steps included in the test run](https://docs.tofupilot.com/steps). Default is None.
            sub_units (Optional[List[SubUnit]], optional): [A list of sub-units included in the test run](https://docs.tofupilot.com/sub-units). Default is None.
            report_variables (Optional[Dict[str, str]], optional): [A dictionary of key values that will replace the procedure's {{report_variables}}](https://docs.tofupilot.com/report). Default is None.
            attachments (Optional[List[str]], optional): [A list of file paths for attachments to include with the test run](https://docs.tofupilot.com/attachments). Default is None.

        Returns:
            dict: A dictionary containing the following keys:
                - success (bool): Whether the test run creation was successful.
                - message (Optional[dict]): Contains URL if successful.
                - status_code (Optional[int]): HTTP status code of the response.
                - error (Optional[dict]): Error message if any.

        Raises:
            requests.exceptions.HTTPError: If the HTTP request returned an unsuccessful status code.
            requests.RequestException: If a network error occurred.
            Exception: For any other exceptions that might occur.

        """
        self._logger.info("Run creation started...")

        if attachments is not None:
            validate_attachments(
                self._logger,
                attachments,
                self._max_attachments,
                self._max_file_size,
            )

        payload = {
            "procedure_id": procedure_id,
            "unit_under_test": unit_under_test,
            "run_passed": run_passed,
            "client": "Python",
            "client_version": self._current_version,
        }

        if steps is not None:
            for step in steps:
                if step["duration"] is not None:
                    step["duration"] = timedelta_to_iso(step["duration"])
                if step["started_at"] is not None:
                    step["started_at"] = datetime_to_iso(step["started_at"])
            payload["steps"] = steps

        if duration is not None:
            payload["duration"] = timedelta_to_iso(duration)

        if sub_units is not None:
            payload["sub_units"] = sub_units

        if report_variables is not None:
            payload["report_variables"] = report_variables

        try:
            response = requests.post(
                f"{self._base_url}/runs",
                json=payload,
                headers=self._headers,
                timeout=SECONDS_BEFORE_TIMEOUT,
            )
            response.raise_for_status()
            json_response = response.json()

            warnings: Optional[List[str]] = json_response.get("warnings")
            if warnings:
                for warning in warnings:
                    self._logger.warning(warning)

            message = json_response.get("message")
            self._logger.success(message)

            run_id = json_response.get("id")

            try:
                if attachments:
                    handle_attachments(
                        self._logger, self._headers, self._base_url, attachments, run_id
                    )
            except Exception as e:
                self._logger.error(e)
                return {
                    "success": False,
                    "message": None,
                    "status_code": None,
                    "error": {"message": str(e)},
                }
            return {
                "success": True,
                "message": message,
                "status_code": response.status_code,
                "error": None,
            }
        except requests.exceptions.HTTPError as http_err:
            error_message = parse_error_message(http_err.response)
            self._logger.error(error_message)
            return {
                "success": False,
                "message": None,
                "status_code": http_err.response.status_code,
                "error": {"message": error_message},
            }
        except requests.RequestException as e:
            self._logger.error("Network error: %s", e)
            return {
                "success": False,
                "message": None,
                "status_code": None,
                "error": {"message": str(e)},
            }
        except Exception as e:
            error_message = f"Failed to create test run: {e}"
            self._logger.error(error_message)
            return {
                "success": False,
                "message": None,
                "status_code": None,
                "error": {"message": error_message},
            }


def print_version_banner(current_version: str):
    """Prints current version of client"""
    banner = f"""
    TofuPilot Python Client {current_version}
    """
    print(banner.strip())



import warnings
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as get_version

import requests
from packaging import version
from ..constants import SECONDS_BEFORE_TIMEOUT


def check_latest_version(logger, current_version, package_name: str):
    """Checks if the package is up-to-date and emits a warning if not"""
    try:
        response = requests.get(
            f"https://pypi.org/pypi/{package_name}/json", timeout=SECONDS_BEFORE_TIMEOUT
        )
        response.raise_for_status()
        latest_version = response.json()["info"]["version"]

        try:
            if version.parse(current_version) < version.parse(latest_version):
                warning_message = (
                    f"You are using {package_name} version {current_version}, however version {latest_version} is available. "
                    f'You should consider upgrading via the "pip install --upgrade {package_name}" command.'
                )
                warnings.warn(warning_message, UserWarning)
                logger.warning(warning_message)
        except PackageNotFoundError:
            logger.info(f"Package {package_name} is not installed.")

    except requests.RequestException as e:
        logger.warning(f"Error checking the latest version: {e}")
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")



"""
This module initializes the TofuPilot package.

It imports the TofuPilotClient class which provides
the main interface for interacting with the TofuPilot API.
"""

from .client import TofuPilotClient



import json
import mimetypes
import os
from datetime import timedelta, datetime
from typing import List, Tuple
import requests

from ..constants.requests import SECONDS_BEFORE_TIMEOUT


def validate_attachments(
    logger,
    attachments: List[str],
    max_attachments: int,
    max_file_size: int,
):
    """Validates a list of attachments by making sure they have the right size"""
    logger.info("Validating attachments...")
    if len(attachments) > max_attachments:
        log_and_raise(
            logger,
            f"Number of attachments exceeds the maximum allowed limit of {max_attachments}",
        )

    for file_path in attachments:
        file_size = os.path.getsize(file_path)
        if file_size > max_file_size:
            log_and_raise(
                logger,
                f"File size exceeds the maximum allowed size of {max_file_size} bytes: {file_path}",
            )


def initialize_upload(headers: dict, base_url: str, file_path: str) -> Tuple[str, str]:
    """Creates a new upload in TofuPilot"""
    initialize_url = f"{base_url}/uploads/initialize"
    file_name = os.path.basename(file_path)
    payload = {"name": file_name}
    response = requests.post(
        initialize_url,
        data=json.dumps(payload),
        headers=headers,
        timeout=SECONDS_BEFORE_TIMEOUT,
    )
    response.raise_for_status()
    response_json = response.json()
    return response_json.get("uploadUrl"), response_json.get("id")


def upload_file(upload_url: str, file_path: str) -> bool:
    """Stores a file into an upload"""
    with open(file_path, "rb") as file:
        content_type, _ = mimetypes.guess_type(file_path) or "application/octet-stream"
        upload_response = requests.put(
            upload_url,
            data=file,
            headers={"Content-Type": content_type},
            timeout=SECONDS_BEFORE_TIMEOUT,
        )
        return upload_response.status_code == 200


def notify_server(headers: dict, base_url: str, upload_id: str, run_id: str) -> bool:
    """Tells TP server to sync upload with newly created run"""
    sync_url = f"{base_url}/uploads/sync"
    sync_payload = {"upload_id": upload_id, "run_id": run_id}
    sync_response = requests.post(
        sync_url,
        data=json.dumps(sync_payload),
        headers=headers,
        timeout=SECONDS_BEFORE_TIMEOUT,
    )
    return sync_response.status_code == 200


def handle_attachments(
    logger, headers: dict, base_url: str, attachments: List[str], run_id: str
):
    """Creates one upload per file and stores them into TofuPilot"""
    for file_path in attachments:
        logger.info(f"Uploading {file_path}...")
        try:
            upload_url, upload_id = initialize_upload(headers, base_url, file_path)
            if upload_url and upload_file(upload_url, file_path):
                if not notify_server(headers, base_url, upload_id, run_id):
                    logger.error(
                        f"Notification Failure: The server could not be notified of the upload for attachment '{file_path}'. The upload may not be recorded in the system. Please check the server status and retry. For more details about run attachments, visit: https://docs.tofupilot.com/attachments."
                    )
                    break
            else:
                logger.error(
                    f"Upload Failure: The attachment '{file_path}' could not be uploaded. This might be due to an inaccessible file path or an issue with the provided upload URL. Verify the file path, ensure your network connection is stable, and try again. For more detail about run attachments, visit: https://docs.tofupilot.com/attachments."
                )
                break
        except requests.RequestException as e:
            logger.error(
                f"Network Error: A network issue occurred while attempting to upload the attachment '{file_path}'. Error details: {e}. Please check your network connection and retry the operation."
            )
            break
        except Exception as e:
            logger.error(
                f"Unexpected Error: An unexpected issue occurred during the upload of the attachment '{file_path}'. Error details: {e}. Please retry the operation."
            )
            break
        logger.success(
            f"Attachment {file_path} successfully uploaded and linked to run."
        )


def parse_error_message(response: requests.Response) -> str:
    try:
        error_data = response.json()
        return error_data.get("error", {}).get(
            "message", f"HTTP error occurred: {response.text}"
        )
    except ValueError:
        return f"HTTP error occurred: {response.text}"


def timedelta_to_iso(td: timedelta) -> str:
    if td == timedelta():  # Check if timedelta is zero
        return "PT0S"  # Return "PT0S" for zero duration

    days = td.days
    seconds = td.seconds
    microseconds = td.microseconds

    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    iso_duration = "P"
    if days > 0:
        iso_duration += f"{days}D"

    if hours > 0 or minutes > 0 or seconds > 0 or microseconds > 0:
        iso_duration += "T"
        if hours > 0:
            iso_duration += f"{hours}H"
        if minutes > 0:
            iso_duration += f"{minutes}M"
        if seconds > 0 or microseconds > 0:
            iso_duration += f"{seconds}"
            if microseconds > 0:
                iso_duration += f".{microseconds:06d}"
            iso_duration += "S"

    return iso_duration


def datetime_to_iso(dt: datetime):
    # Note: using dt.isoformat() does not add the trailing 'Z' which prevents the iso string from being recognized by TP' API
    return dt.strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def log_and_raise(logger, error_message: str):
    logger.error(error_message)
    raise RuntimeError(error_message)



from .logger import setup_logger
from .version_checker import check_latest_version
from .file_utils import (
    validate_attachments,
    initialize_upload,
    upload_file,
    notify_server,
    handle_attachments,
    parse_error_message,
    timedelta_to_iso,
    datetime_to_iso,
    log_and_raise,
)

__all__ = [
    "setup_logger",
    "check_latest_version",
    "validate_attachments",
    "initialize_upload",
    "upload_file",
    "notify_server",
    "handle_attachments",
    "parse_error_message",
    "timedelta_to_iso",
    "datetime_to_iso",
    "log_and_raise",
]



from setuptools import setup, find_packages

with open("README.md", "r") as file:
    long_description = file.read()

setup(
    name="tofupilot",
    version="1.2.0",
    packages=find_packages(),
    install_requires=["requests", "setuptools", "packaging"],
    author="Félix Berthier",
    author_email="felix.berthier@tofupilot.com",
    description="The official Python client for the TofuPilot API",
    license="MIT",
    keywords="automatic hardware testing tofupilot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tofupilot/python-client",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
